from unittest import TestCase

from baca2PackageManager.tools import bytes_to_str


class ToolsTest(TestCase):
    def test_bytes_to_str(self):
        self.assertEqual(bytes_to_str(1), '1B')
        self.assertEqual(bytes_to_str(512), '512B')
        self.assertEqual(bytes_to_str(1024), '1.00K')
        self.assertEqual(bytes_to_str(1536), '1.50K')
        self.assertEqual(bytes_to_str(1024 ** 2), '1.00M')
        self.assertEqual(bytes_to_str(1024 ** 3), '1.00G')

    def test_bytes_to_str_zero(self):
        self.assertEqual(bytes_to_str(0), '0B')


from __future__ import annotations
from typing import Self, Any

from enum import Enum
from abc import ABC, abstractmethod

import pickle


class JudgeVerdict(Enum):
    """Judge Verdict"""
    OK = 0
    FAIL = 1
    INCONCLUSIVE = 2


class JudgeNodeBase(ABC):
    """
    Abstract base for 'Judge Nodes' of decision graphs in 'JudgeMaster' class.

    A 'JudgeNode' is supposed to represent a single step in a process of submit checking.
    It should implement a specific test for submits and return a verdict in form of the
    'JudgeVerdict' class instance.
    """

    def __init__(self, name: str):
        """
        Initiate JudgeNodeBase.

        :param name: a unique identifier
        :type name: str
        """
        self._id = name

    def __hash__(self):
        return hash(self.name)

    def __eq__(self, other):
        if not isinstance(other, JudgeNodeBase):
            return False
        return self.name == other.name

    @property
    def name(self) -> str:
        """
        Returns a string used for identification purposes

        :return: a string used for identification purposes
        """
        return self._id

    @abstractmethod
    def start(self, *args, **kwargs) -> Any:
        """
        Starts judging process.
        This method should initiate checking process for a submit which in the end
        should call the 'receive' method of the same instance.
        """
        ...

    @abstractmethod
    def receive(self, *args, **kwargs) -> JudgeVerdict:
        """
        Receives judging results and returns a verdict.
        This method should be called after the 'start' method and the process it
        had initiated.

        :return: a verdict used for traversing the decision graph in JudgeMaster
        """
        ...

    def serialise(self) -> bytes:
        return pickle.dumps(self)

    @classmethod
    def unpack(cls, binary_data: bytes) -> Self:
        tmp = pickle.loads(binary_data)
        if not isinstance(tmp, cls):
            raise TypeError(f"Unpacked object is not of type {cls.__name__}")
        return tmp


class EndNode(JudgeNodeBase):
    """
    Special type of the 'JudgeNodeBase' class. Represents the final stage of submit checking
    process and yields the final verdict of all tests.
    """

    def __init__(self, name: str, end_node_meaning: JudgeVerdict):
        """
        Initiate an instance

        :param name: unique identifier
        :type name: str
        :param end_node_meaning: the final verdict of all tests
        :type end_node_meaning: JudgeVerdict
        """
        super().__init__(name)
        self.meaning = end_node_meaning

    def start(self, *args, **kwargs) -> Any:
        pass

    def receive(self, *args, **kwargs) -> JudgeVerdict:
        """
        Returns the final verdict. Ignores all arguments.

        :return: the final verdict
        """
        return self.meaning


class JudgeNodeError(ValueError):
    """Exception class used by JudgeMaster"""
    pass


class EndNodeError(JudgeNodeError):
    """Exception class used by JudgeMaster"""
    pass


class JudgeGraphIntegrityReport:
    """
    Data class used to hold information about errors in the decision graph in
    'JudgeMaster' class.

    attr unreachable_nodes: stores nodes to which no path from the start_node exists
    attr cannot_reach_end: stores nodes from which no path leads to any EndNode
    attr wrong connections: stores nodes for which edges are defined in a wrong way
    attr has_end_nodes: True if the decision graph has any end nodes, False otherwise
    """
    def __init__(self,
                 unreachable_nodes: list[JudgeNodeBase],
                 cannot_reach_end: list[JudgeNodeBase],
                 wrong_connections: list[JudgeNodeBase],
                 has_end_nodes: bool):
        self.unreachable_nodes = unreachable_nodes
        self.cannot_reach_end = cannot_reach_end
        self.wrong_connections = wrong_connections
        self.has_end_nodes = has_end_nodes

    @property
    def no_errors(self) -> bool:
        """Returns True if the decision graph is completely error free"""
        return not (
                self.unreachable_nodes or
                self.cannot_reach_end or
                self.wrong_connections or
                not self.has_end_nodes)


class JudgeManager:
    """Manages the decision graph of submit checking"""

    def __init__(self):
        self.graph: dict[JudgeNodeBase, dict[JudgeVerdict, JudgeNodeBase]] = {}
        self._start_node: JudgeNodeBase = None

    def __eq__(self, other):
        if not isinstance(other, JudgeManager):
            return False
        return self.graph == other.graph and self._start_node == other._start_node

    @property
    def nodes(self) -> list[JudgeNodeBase]:
        """:return: a list of all nodes in the graph"""
        return list(self.graph.keys())

    def get_node_by_name(self, name: str) -> JudgeNodeBase:
        """
        Looks for a node with name 'name'.

        :param name: name of the wanted node
        :return: the node with name 'name'
        :raise JudgeNodeError: if there is no node with name 'name'
        """
        for node in self.nodes:
            if node.name == name:
                return node
        else:
            raise JudgeNodeError(f"No node with name {name}.")

    def set_start_node(self, node: JudgeNodeBase):
        """Set start node. It is None by default"""
        if node not in self.graph:
            raise KeyError(repr(node))
        self._start_node = node

    def get_start_node(self) -> JudgeNodeBase:
        """
        Get start node.

        :return: the start node
        :raise JudgeNodeError: if start node is not set
        """
        if self._start_node is None:
            raise JudgeNodeError("Start_node is not set.")
        return self._start_node

    @classmethod
    def from_dict(cls,
                  dct: dict[JudgeNodeBase, dict[JudgeVerdict, JudgeNodeBase]],
                  start_node: JudgeNodeBase) -> Self:
        """
        Create a JudgeManager class instance using a dictionary.

        :param dct: a graph in form of an adjacency list
        :param start_node: the start node
        :return: a JudgeMaster class instance with the provided graph
        """
        out = cls()
        for node in dct:
            out.add_node(node)
        for from_node, adj_list in dct.items():
            for verdict, to_node in adj_list.items():
                out.add_connection(from_node, to_node, verdict)
        out.set_start_node(start_node)
        return out

    def add_node(self, node: JudgeNodeBase):
        """
        Add node to the graph. node.name has to be unique.

        :param node: the node to be added.
        :raise JudgeNodeError: if the node already belongs to the graph
        """
        if node in self.graph:
            raise JudgeNodeError(f"Node with name '{node.name}' has already been added.")
        self.graph[node] = {}

    def add_connection(self, from_node: JudgeNodeBase, to_node: JudgeNodeBase, with_verdict: JudgeVerdict):
        """
        Create a (directed) edge between to nodes and label it with verdict. If an edge with the
        provided verdict already exists it will be overwritten.

        :param from_node: the edge will begin at this node
        :param to_node: the edge will end at this node
        :param with_verdict: the edge will be labeled wit this verdict

        :raise KeyError: if from_node or to_node do not belong to the graph
        :raise ResourceWarning: if trying to add edges from EndNodes
        """
        if isinstance(from_node, EndNode):
            raise ResourceWarning(f"Connections from EndNodes will never be used.")
        adj_list = self.graph[from_node]
        if to_node not in self.graph:
            raise KeyError(repr(to_node))
        adj_list[with_verdict] = to_node

    def remove_node(self, node: JudgeNodeBase):
        """
        Remove the provided node and all edges leaving or entering this node from the graph.

        :param node: the node to be removed
        :raise KeyError: if the node does not belong to the graph
        """
        del self.graph[node]
        for key in self.graph:
            adj_list = self.graph[key]
            new_adj_list = {a: b for a, b in adj_list.items() if b is not node}
            self.graph[key] = new_adj_list

    def remove_connection_by_node(self, from_node: JudgeNodeBase, to_node: JudgeNodeBase):
        """
        Remove all edges between two specified nodes.

        :param from_node: the first node
        :param to_node: the last node
        :raise KeyError: if any of the nodes do not belong to the graph or there are no edges between them
        """
        adj_list = self.graph[from_node]
        tmp: list[JudgeNodeBase] = []
        for key in adj_list:
            if adj_list[key] is to_node:
                tmp.append(key)
        if tmp:
            for item in tmp:
                del adj_list[item]
        else:
            raise KeyError("No connections")

    def remove_connection_by_verdict(self, from_node: JudgeNodeBase, verdict: JudgeVerdict):
        """
        Remove an edge leaving from from_node labeled with the specified verdict

        :param from_node: the node for which an edge should be deleted
        :param verdict: label of the edge that should be deleted
        :raise KeyError: if the node or the edge do not exist
        """
        adj_list = self.graph[from_node]
        del adj_list[verdict]

    def send(self, node: JudgeNodeBase, *args, **kwargs) -> Any:
        """
        Invokes the 'send' method on the provided node with provided arguments.

        :param node: node to be used
        :raise KeyError: if the given node does not belong to the graph
        """
        if node not in self.graph:
            raise KeyError(repr(node))
        return node.start(*args, **kwargs)

    def advance(self, node: JudgeNodeBase | str, verdict: JudgeVerdict) -> JudgeNodeBase | None:
        """
        Advances the decision graph according to the provided verdict.

        :return: Next node according to the graph, if there is no edge for provided verdict: None
        :raise KeyError: if the given node does not belong to the graph
        :raise EndNodeError: if the given node is an EndNode
        """
        if isinstance(node, str):
            node = self.get_node_by_name(node)

        if isinstance(node, EndNode):
            raise EndNodeError("'node' cannot be an EndNode")
        return self.graph[node].get(verdict)

    def check_graph_integrity(self) -> JudgeGraphIntegrityReport:
        """
        Looks for errors in the decision graph and returns a report.

        :return: the report
        """
        reach_list = self._floyd_warshall()
        end_nodes = frozenset(node for node in self.nodes if isinstance(node, EndNode))

        # unreachable nodes searching
        unreachable: list[JudgeNodeBase] = list(frozenset(self.nodes).difference(reach_list[self._start_node]))

        # searching for nodes from which no EndNode is reachable
        cannot_reach_end: list[JudgeNodeBase] = []
        for node in self.nodes:
            if not end_nodes.intersection(reach_list[node]):
                cannot_reach_end.append(node)

        # checking for valid connections
        wrong_connections: list[JudgeNodeBase] = []
        for node in self.nodes:
            tmp = frozenset(self.graph[node].keys())
            if not (
                    {JudgeVerdict.OK, JudgeVerdict.FAIL}.issubset(tmp)
                    or {JudgeVerdict.OK, JudgeVerdict.INCONCLUSIVE}.issubset(tmp)
                    or isinstance(node, EndNode)
            ):
                wrong_connections.append(node)

        # check for end nodes
        has_end_nodes = len(end_nodes) != 0

        return JudgeGraphIntegrityReport(unreachable, cannot_reach_end, wrong_connections, has_end_nodes)

    def _floyd_warshall(self) -> dict[JudgeNodeBase, frozenset[JudgeNodeBase]]:
        node_list = self.nodes  # for performance
        nl_length = len(node_list)
        visit_matrix = [[j in self.graph[i].values() or i is j for j in node_list] for i in node_list]

        for k in range(nl_length):
            for i in range(nl_length):
                for j in range(nl_length):
                    visit_matrix[i][j] = visit_matrix[i][j] or (visit_matrix[i][k] and visit_matrix[k][j])

        out: dict[JudgeNodeBase, list[JudgeNodeBase]] = {}
        for index, node in enumerate(node_list):
            out[node] = frozenset(node_list[i] for i, b in enumerate(visit_matrix[index]) if b)

        return out

    def serialise(self) -> bytes:
        return pickle.dumps(self)

    @classmethod
    def unpack(cls, binary_data: str) -> Self:
        tmp = pickle.loads(binary_data)
        if not isinstance(tmp, cls):
            raise TypeError(f"Unpacked object is not of type {cls.__name__}")
        return tmp



from pathlib import Path
from re import findall, split

from baca2PackageManager.tools import bytes_from_str


def isAny(val):
    """
    any non-empty value is allowed

    :return: A boolean value.
    """
    return bool(val)


def isNone(val):
    """
    check if val is None

    :return: A boolean value.
    """
    return val is None


def isInt(val):
    """
    check if val can be converted to int

    :return: A boolean value.
    """
    if type(val) == float:
        return False
    try:
        int(val)
        return True
    except ValueError:
        return False


def isIntBetween(val, a: int, b: int):
    """
    check if val is an int value between a and b

    :return: A boolean value.
    """
    if isInt(val):
        if a <= val < b:
            return True
    return False


def isFloat(val):
    """
    check if val can be converted to float

    :return: A boolean value.
    """
    try:
        float(val)
        return True
    except ValueError:
        return False


def isFloatBetween(val, a: int, b: int):
    """
    check if val is a float value between a and b

    :return: A boolean value.
    """
    if isFloat(val):
        if a <= val < b:
            return True
    return False


def isStr(val):
    """
    check if val can be converted to string

    :return: A boolean value.
    """
    if type(val) == str:
        return True
    return False


def is_(val, schema: str):
    """
    check if val is exactly like schema

    :return: A boolean value.
    """
    if isStr(val):
        return val == schema
    return False


def isIn(val, *args):
    """
    check if val is in args

    :return: A boolean value.
    """
    return val in args


def isBool(val):
    """
    check if val is bool

    :return: A boolean value.
    """
    return type(val) == bool


def isShorter(val, l: int):
    """
    check if val is string and has len < len(l)

    :return: A boolean value.
    """
    if isStr(val):
        return len(val) < l
    return False


def isDict(val):
    """
    check if val has dict type

    :return: A boolean value.
    """
    return type(val) == dict


def isPath(val):
    """
    check if val is path in package_dir

    :return: A boolean value.
    """
    if val is None:
        return False
    elif isinstance(val, Path) and val.exists():
        return True
    try:
        val = Path(val)
        if val.exists():
            return True
        return False
    except ValueError:
        return False


def resolve_validator(func_list, arg):
    """
    takes the validator function with arguments, and check that if validator function is true for arg (other arguments for func)
    """
    func_name = str(func_list[0])
    func_arguments_ext = ',' + ','.join(func_list[1:])
    return eval(func_name + '("' + str(arg) + '"' + func_arguments_ext + ')')


def hasStructure(val, struct: str):
    """
    check if val has structure provided by struct and fulfills validators functions from struct

    :return: A boolean value.
    """
    validators = findall("<.*?>", struct)
    validators = [i[1:-1].split(',') for i in validators]
    constant_words = findall("[^<>]{0,}<", struct) + findall("[^>]{0,}$", struct)
    constant_words = [i.strip("<") for i in constant_words]
    if len(validators) == 1:
        values_to_check = [val]
    else:
        # words_in_pattern = [i for i in constant_words if i != '|' and i != '']
        # regex_pattern = '|'.join([i for i in constant_words if i != '|' and i != ''])
        values_to_check = split('|'.join([i for i in constant_words if i != '|' and i != '']), val)
    if struct.startswith('<') == False:
        values_to_check = values_to_check[1:]
    valid_idx = 0
    const_w_idx = 0
    values_idx = 0
    temp_alternative = False
    result = True
    while valid_idx < len(validators) and values_idx < len(values_to_check):
        temp_alternative |= resolve_validator(validators[valid_idx], values_to_check[values_idx])
        if constant_words[const_w_idx] == '|':
            if constant_words[const_w_idx + 1] != '|':
                values_idx += 1
        else:
            if constant_words[const_w_idx + 1] != '|':
                values_idx += 1
                result &= temp_alternative
                temp_alternative = False
        valid_idx += 1
        const_w_idx += 1
    return result


def valid_memory_size(first: str, second: str):
    """
    checks if first is smaller than second considering memory

    :return: A boolean value.
    """
    if bytes_from_str(first) <= bytes_from_str(second):
        return True
    return False


def isSize(val: str, max_size: str):
    """
    check if val has structure like <isInt><isIn, 'B', 'K', 'M', 'G'>

    :return: A boolean value.
    """
    val = val.strip()
    val.replace(' ', '')
    return all((
        hasStructure(val[:-1], "<isInt>"),
        hasStructure(val[-1], "<isIn, 'B', 'K', 'M', 'G'>"),
        valid_memory_size(val, max_size)
    ))


def isList(val, *args):
    """
    check if val is a list and every element from list fulfill at least one validator from args

    :return: A boolean value.
    """
    if isinstance(val, list):
        if not args:
            return True

        for i in val:
            if not any([vl[0](i, *vl[1:]) for vl in args]):
                return False

        return True
    return False



from pathlib import Path


class FileAlreadyExist(Exception):
    pass


class NoTestFound(Exception):
    pass


class NoSetFound(Exception):
    pass


class TestExistError(FileAlreadyExist):
    pass


class InvalidFileExtension(Exception):
    pass


class PackageCreationFailed(Exception):
    def __init__(self, e, **kwargs):
        super().__init__(self.message(e, **kwargs))

    @staticmethod
    def message(e: Exception, **kwargs) -> str:
        msg = 'Package creation failed: '
        if isinstance(e, FileNotFoundError):
            filename = str(e).split()[-1].replace('\'', '')
            path = Path(filename)
            msg += f'File \'{path.relative_to(kwargs.get("root_path"))}\' not found'
        else:
            msg += str(e)
        return msg



import shutil
from pathlib import Path
from typing import Iterable

from importlib.resources.abc import StrPath
from zipfile import ZipFile, ZipInfo


class Zip(ZipFile):
    class UnsafeZipFile(Exception):
        """
        Exception raised when the zip file is unsafe.
        """
        pass

    ZIPBOMB_RATIO = 20

    @property
    def uncompressed_size(self) -> int:
        """
        Returns the size of the compressed data.

        :return: Size of the compressed data in bytes [B].
        :rtype: int
        """
        return sum([data.file_size for data in self.filelist])

    @property
    def compressed_size(self) -> int:
        """
        Returns the size of the uncompressed data.

        :return: Size of the uncompressed data in bytes [B].
        :rtype: int
        """
        return sum([data.compress_size for data in self.filelist])

    def check_zip_bomb(self) -> None:
        """
        Checks if the zip bomb is present in the zip file.

        :param max_size: Maximum size of the compressed data in bytes [B].
        :type max_size: int
        :return: True if the zip bomb is present, False otherwise.
        :rtype: bool
        """
        if self.uncompressed_size / self.compressed_size > self.ZIPBOMB_RATIO:
            raise self.UnsafeZipFile('Zip bomb detected.')

    def move_to_top(self, path: StrPath | Path, leave_top: bool = True) -> None:
        """
        Shreds the 1 directory chain from zip file. To be called after zip extraction.

        :return: None
        """
        path = Path(path) if isinstance(path, str) else path

        path_glob = list(path.glob('*'))
        if len(path_glob) == 1:
            sub_path = path_glob[0]
            if leave_top:
                self.move_to_top(sub_path, False)
                return
            mv_path = path.parent / f'{path.name}_bak'
            shutil.rmtree(mv_path, ignore_errors=True)
            shutil.move(sub_path, mv_path)
            shutil.rmtree(path, ignore_errors=True)
            mv_path.rename(path)

            # org_name = path.name
            # path.rename(path.parent / f'{org_name}_bak')
            # sub_path = list(path.glob('*'))[0]
            # sub_path.rename(path.parent / org_name)
            # path.rmdir()
            self.move_to_top(path, False)

    def extractall(
            self,
            path: StrPath | None = ...,
            members: Iterable[str | ZipInfo] | None = None,
            pwd: bytes | None = ...,
            leave_top: bool = True
    ):
        """
        Extract all members from the archive to the current working
        directory. `path' specifies a different directory to extract to.
        `members' is optional and must be a subset of the list returned
        by namelist().
        """
        self.check_zip_bomb()
        super().extractall(path, members, pwd)
        self.move_to_top(path, leave_top)

    def check_extensions(self, allowed_extensions: Iterable[str]):
        """
        Checks if the zip file contains only files with allowed extensions.

        :param allowed_extensions: Allowed file extensions.
        :type allowed_extensions: Iterable[str]
        :return: True if the zip file contains only files with allowed extensions, False otherwise.
        :rtype: bool
        """
        for file in self.filelist:
            if Path(file.filename).suffix[1:].lower() not in allowed_extensions:
                return False
        return True




from __future__ import annotations

import shutil
from copy import deepcopy
from enum import Enum
from pathlib import Path
from os import remove, replace, walk, mkdir, rename, listdir
from shutil import rmtree, copytree
from typing import List, Self

from yaml import safe_load, dump
from re import match

from .judge_manager import JudgeManager
from .validators import isAny, isNone, isInt, isIntBetween, isFloat, isFloatBetween, isStr, is_, \
    isIn, isShorter, \
    isDict, isPath, isSize, isList, valid_memory_size, isBool
from .tools import bytes_from_str
from .consts import SUPPORTED_EXTENSIONS, BASE_DIR
from .manager_exceptions import NoTestFound, NoSetFound, TestExistError, FileAlreadyExist, \
    InvalidFileExtension, PackageCreationFailed

__all__ = ['Package', 'TSet', 'TestF']

from .zipper import Zip


def merge_settings(default: dict, to_add: dict) -> dict:
    """
    It takes two dictionaries, and returns a new dictionary that has the keys of the first dictionary, and the values of the
    second dictionary if they exist, otherwise the values of the first dictionary. It overwrites default dict with valuses from to_add

    :param default: The default settings
    :type default: dict
    :param to_add: The settings you want to add to the default settings
    :type to_add: dict

    :return: A dictionary with the keys of the default dictionary and the values of the to_add dictionary.
    """
    new = {}
    for i in default.keys():
        if to_add is not None:
            if i in to_add.keys() and to_add[i] is not None:
                new[i] = to_add[i]
            else:
                new[i] = default[i]
        else:
            new[i] = default[i]
    return new


class PackageManager:
    """
    It's a class that manages a package's settings
    """

    def __init__(self, path: Path, settings_init: Path or dict, default_settings: dict):
        """
        If the settings_init is a dict, assign it to settings. If not, load the settings_init yaml file and assign it to
        settings. Then merge the settings with the default settings

        :param path: Path to the settings file
        :type path: Path
        :param settings_init: This is the path to the settings file
        :type settings_init: Path or dict
        :param default_settings: a dict with default settings
        :type default_settings: dict
        """
        self._path = path
        # if type of settings_init is a dict assign settings_init to processed settings
        settings = {}
        if type(settings_init) == dict:
            if bool(settings_init):
                settings = settings_init
        # if not, make dict from settings_init yaml
        else:
            # unpacking settings file to dict
            with open(settings_init, mode="rt", encoding="utf-8") as file:
                settings = safe_load(file)
        # merge external settings with default
        self._settings = merge_settings(default_settings, settings)

    def __getitem__(self, arg: str):
        """
        If the key is in the dictionary, return the value. If not, raise a KeyError

        :param arg: The name of the key to get the value of
        :type arg: str

        :return: The value of the key in the dictionary.
        """
        try:
            return self._settings[arg]
        except KeyError:
            raise KeyError(f'No key named {arg} has found in self_settings')

    def get(self, arg: str, default=None):
        try:
            return self._settings[arg]
        except KeyError:
            return default

    def __setitem__(self, arg: str, val):
        """
        `__setitem__` is a special method that allows us to use the `[]` operator to set a value in a dictionary

        :param arg: str
        :type arg: str
        :param val: the value to be set
        """
        self._settings[arg] = val
        # effect changes to yaml settings
        self.save_to_config(self._settings)

    def check_validation(self, validators) -> bool:
        """
        It checks if the value of the setting is valid by checking if it matches any of the validators for that setting

        :param validators: A dictionary of validators. The keys are the names of the settings, and the values are lists of
        validators. Each validator is a tuple of the form (function, *args, **kwargs). The function is called with the
        setting value as the first argument, followed by the *

        :return: The check variable is being returned.
        """
        return True
        # for i, j in self._settings.items():
        #     check = False
        #     for k in validators[i]:
        #         check |= k[0](j, *k[1:])
        #     if not check:
        #         return False
        # return True

    def save_to_config(self, settings):
        """
        It opens a file called config.yml in the directory specified by the path attribute of the object, and writes the
        settings dictionary to it.

        :param settings: The settings to save
        """
        with open(self._path / 'config.yml', mode="wt", encoding="utf-8") as file:
            dump(settings, file)

    def read_from_config(self):
        """
        It reads the config.yml file from the path and returns the contents

        :return: The dict from config.yml file is being returned.
        """
        with open(self._path / 'config.yml', mode="rt", encoding="utf-8") as file:
            return safe_load(file)

    def add_empty_file(self, filename):
        """
        It creates an empty file if it doesn't already exist

        :param filename: The name of the file to be created
        """
        if not isPath(self._path / filename):
            with open(self._path / filename, 'w') as f:
                pass

    def __iter__(self):
        """
        It returns an iterator for the settings dictionary

        :return: The iterator for the settings dictionary
        """
        return iter(self._settings.items())


class Package(PackageManager):
    """
    It's a class that represents a package.
    """

    #: Largest file acceptable to upload
    MAX_SUBMIT_MEMORY = '10G'
    #: Largest acceptable submit time
    MAX_SUBMIT_TIME = 600
    MAX_CPUS = 16
    MAX_SOURCE_SIZE = '512M'
    SETTINGS_VALIDATION = {
        'title': [[isStr]],
        'points': [[isInt], [isFloat]],
        'memory_limit': [[isSize, MAX_SUBMIT_MEMORY]],
        'source_memory': [[isSize, MAX_SOURCE_SIZE]],
        'time_limit': [[isIntBetween, 0, MAX_SUBMIT_TIME], [isFloatBetween, 0, MAX_SUBMIT_TIME]],
        'allowedExtensions': [[isIn, *SUPPORTED_EXTENSIONS],
                              [isList, *SUPPORTED_EXTENSIONS]],
        'hinter': [[isNone], [isPath]],
        'checker': [[isNone], [isPath]],
        'verifier': [[isNone], [isPath]],
        'basename': [[isNone], [isStr]],
        'test_generator': [[isNone], [isPath]],
        'network': [[isNone], [isBool]],
        'cpus': [[isNone], [isIntBetween, 1, MAX_CPUS]]
    }
    """
    Validation for ``Package`` settings.
    
    Available options are:
    
        * ``title``: package name
        * ``points``: maximum amount of points to earn
        * ``memory_limit``: is a memory limit
        * ``time_limit``: is a time limit
        * ``allowedExtensions``: extensions witch are accepted
        * ``hinter``: is a path or None value to actual hinter
        * ``checker``: is a path or None value to actual checker
        * ``test_generator``: is a path or None value to actual generator
        * ``network``: is a bool value to allow network access
        * ``cpus``: is a number of cpus to use
    """

    #: Default values for Package settings
    DEFAULT_SETTINGS = {
        'title': '<no-name>',
        'points': 0,
        'memory_limit': '512M',
        'time_limit': 10,
        'allowedExtensions': 'cpp',
        'hinter': None,
        'checker': None,
        'test_generator': None,
        'network': False,
        'verifier': None,
        'basename': None,
        'cpus': 1
    }

    # TODO: Add auto-discovery of judge manager files

    class DocExtension(Enum):
        """
        Defines the allowed extensions for task description files
        """
        HTML = 'html'
        MD = 'md'
        PDF = 'pdf'
        TXT = 'txt'

        @classmethod
        def allowed_extensions(cls) -> List[str]:
            """
            :return: List of allowed extensions
            :rtype: List[str]
            """
            return [e.value for e in cls]

        @classmethod
        def from_str(cls, ext: str) -> Self:
            """
            Creates an instance of DocExtension from string, checking if the extension is valid

            :param ext: The extension
            :type ext: str
            :return: Instance of DocExtension
            :rtype: DocExtension

            :raise InvalidFileExtension: If the extension is not valid
            """
            if ext.lower() not in cls.allowed_extensions():
                raise InvalidFileExtension(f'"{ext}" is not valid extension for task description')
            return cls[ext.upper()]

    def __init__(self, path: Path, commit: str, validate_pkg: bool = False) -> None:
        """
        It takes a path to a folder, and then it creates a list of all the subfolders in that folder, and then it creates a
        TSet object for each of those subfolders

        :param path: Path - the path to the package
        :type path: Path
        """
        self._path = path
        self._commit = commit

        config_path = self.commit_path / 'config.yml'
        if not config_path.is_file():
            config_path = self.commit_path / 'config.yaml'
        sets_path = self.commit_path / 'tests'
        super().__init__(path, config_path, Package.DEFAULT_SETTINGS)
        self._sets = []
        self.judge_manager = None
        for i in [x[0].replace(str(sets_path) + '\\', '') for x in walk(sets_path)][1:]:
            self._sets.append(TSet(sets_path / i, inherit_settings=self._settings))
        if isinstance(self['allowedExtensions'], str):
            self['allowedExtensions'] = [self['allowedExtensions']]
        if validate_pkg:
            self.check_package()

    @classmethod
    def create_from_zip(cls,
                        path: Path,
                        commit: str,
                        zip_path: Path,
                        overwrite: bool = False) -> Self:
        """
        It takes a path to a zip file, and then it extracts it to a folder, and then it creates a Package object from that
        folder

        :param path: Path - the path to the zip file
        :type path: Path
        :param commit: The commit of the package
        :type commit: str
        :param zip_path: The path to the zip file
        :type zip_path: Path
        :param overwrite: If True, it will overwrite the package if it already exists,
            defaults to False
        :type overwrite: bool, optional

        :return: The Package object
        :rtype: Package
        """
        pkg_path = path / commit
        if pkg_path.is_dir():
            if overwrite:
                rmtree(pkg_path)
            else:
                raise FileAlreadyExist(f'Files already exists in {pkg_path}')
        pkg_path.mkdir(parents=True)
        try:
            with Zip(zip_path, 'r') as zip_f:
                zip_f.extractall(pkg_path, leave_top=False)

            pkg = cls(path, commit, validate_pkg=True)
        except Exception as e:
            rmtree(pkg_path)
            if not any(path.iterdir()):
                rmtree(path)

            raise PackageCreationFailed(e, root_path=pkg_path)
        return pkg

    @property
    def name(self) -> str:
        """
        :return: It returns the name of the package
        """
        return self._settings['title']

    def set_judge_manager(self, judge_manager: JudgeManager):
        self.judge_manager = judge_manager

    @property
    def judge_manager_path(self) -> Path:
        jm_path = self.commit_path / 'sequence.judge'
        if not jm_path.is_file():
            raise FileNotFoundError(f'Judge manager file not found in {self.commit_path}')
        return jm_path

    def prepare_build(self, build_name: str):
        """
        It prepares the build

        :param build_name: The name of the build
        :type build_name: str
        """
        if not (self.commit_path / '.build').is_dir():
            mkdir(self.commit_path / '.build')

        build_dir = self.commit_path / '.build' / build_name

        if build_dir.is_dir():
            rmtree(build_dir)
        mkdir(build_dir)

        return build_dir

    def build_path(self, build_name: str) -> Path:
        """
        It returns the path to the build

        :param build_name: The name of the build
        :type build_name: str

        :return: The path to the build
        :rtype: Path
        """
        return self.commit_path / '.build' / build_name

    def check_build(self, build_name: str) -> bool:
        """
        It checks if the build exists

        :param build_name: The name of the build
        :type build_name: str

        :return: True if the build exists, False otherwise
        :rtype: bool
        """
        build_dir = self.build_path(build_name)

        if not build_dir.is_dir():
            return False

        return any(list(walk(build_dir))[0][1:])

    def delete_build(self, build_name: str = None):
        """
        Deletes the build. If build_name is None, it deletes all builds.

        :param build_name: The name of the build, if None, it clears all builds
        :type build_name: str
        """
        if build_name is None:
            rmtree(self.commit_path / '.build')
        else:
            rmtree(self.commit_path / '.build' / build_name)

    @property
    def commit_path(self) -> Path:
        """
        It returns the path to the commit
        TODO: After implementing internal git system, this should be changed

        :return: The path to the commit
        :rtype: Path
        """
        return self._path / self._commit

    def rm_tree(self, set_name):
        """
        It removes a directory tree

        :param set_name: The name of the test set
        """
        if isPath(self.commit_path / 'tests' / set_name):
            rmtree(self.commit_path / 'tests' / set_name)

    def make_commit(self, new_commit: str) -> 'Package':
        """
        Function copies the package instance and creates new one/

        :param new_path: The path of the package
        :param new_commit: New unique commit

        :return: new Package
        """
        new_commit_path = self._path / new_commit
        copytree(self.commit_path, new_commit_path)
        return Package(self._path, new_commit)

    def delete(self) -> None:
        """
        Function deletes the package (itself) from directory

        :return: None
        """
        rmtree(self.commit_path)
        del self

    def _add_new_set(self, set_name) -> 'TSet':
        """
        This function adds a new test set to the test suite

        :param set_name: The name of the new test set

        :return: A new TSet object.
        """
        settings = {'name': set_name} | self._settings
        set_path = self.commit_path / 'tests' / set_name
        if not isPath(set_path):
            mkdir(set_path)
            with open(set_path / 'config.yml', 'w') as file:
                dump({'name': set_name}, file)
        new_set = TSet(set_path)
        self._sets.append(new_set)
        return new_set

    def sets(self, set_name: str = None, add_new: bool = False) -> TSet or List[TSet]:
        """
        It returns the set with the name `set_name` if it exists, otherwise it raises an error
        If set_name is None, it returns list of all sets.

        :param set_name: The name of the set you want to get (if none it returns all sets)
        :type set_name: str
        :param add_new: If True, it will create a new set directory if it doesn't exist, defaults to False
        :type add_new: bool (optional)

        :return: The set with the name set_name, or a list of all sets
        :rtype: TSet or List[TSet]
        """
        if set_name is None:
            return self._sets

        for i in self._sets:
            if i['name'] == set_name:
                return i
        if add_new:
            self._add_new_set(set_name)
        else:
            raise NoSetFound(f'Any set directory named {set_name} has found')

    def delete_set(self, set_name: str):
        """
        It deletes a set from the sets list and removes the directory of the set

        :param set_name: The name of the set you want to delete
        :type set_name: str
        :return: the list of sets.
        """
        for i in self._sets:
            if i['name'] == set_name:
                self._sets.remove(i)
                self.rm_tree(set_name)
                return
        raise NoSetFound(f'Any set directory named {set_name} has found to delete')

    def check_package(self, subtree: bool = True) -> bool | int:
        """
        It checks the package.

        :param subtree: bool = True, defaults to True
        :type subtree: bool (optional)
        :return: The result of the check_validation() method and the result of the check_set() method.
        """
        result = True
        # check doc file
        try:
            self.doc_extension()
        except FileNotFoundError:
            return False

        # check sets
        if subtree:
            for i in self._sets:
                result &= i.check_set()

        # check package
        return self.check_validation(Package.SETTINGS_VALIDATION) & result

    def check_source(self, source_code: Path) -> None:
        """
        It checks the source code

        :param source_code: The path to the source code
        :type source_code: Path
        """
        if not source_code.is_file():
            raise FileNotFoundError(f'File not found')
        if source_code.suffix[1:] not in self['allowedExtensions']:
            if source_code.suffix[1:].lower() == 'zip':
                with Zip(source_code) as zip_f:
                    if not zip_f.check_extensions(self['allowedExtensions']):
                        raise InvalidFileExtension('Zipped file contains files with invalid '
                                                   'extension')
            raise InvalidFileExtension(f'Submitted file has invalid extension')

    def doc_path(self, extension: str | DocExtension) -> Path:
        """
        It returns the path to the task description file with the given extension.

        Valid extensions are:

        * PDF
        * MD
        * HTML
        * TXT

        :param extension: The extension of the file
        :type extension: str
        :return: The path to the task description file
        :rtype: Path

        :raise FileNotFoundError: If the file with the given extension doesn't exist
        """
        if isinstance(extension, str):
            extension = self.DocExtension.from_str(extension)
        doc_path = self.commit_path / 'doc'
        path = None
        for f in doc_path.iterdir():
            if f.suffix == f'.{extension.value}':
                path = f
                break
        if not (path and path.is_file()):
            raise FileNotFoundError(f'"{extension.name}" is not valid extension for '
                                    f'{self.get("title")} task description')
        return path

    def doc_has_extension(self, extension: str | DocExtension) -> bool:
        """
        It checks if the task description file has the given extension.

        Valid extensions are:

        * PDF
        * MD
        * HTML
        * TXT

        :param extension: The extension of the file
        :type extension: str
        :return: True if the file with the given extension exists, False otherwise
        :rtype: bool
        """
        try:
            self.doc_path(extension)
            return True
        except FileNotFoundError:
            return False
        except Exception as e:
            print(e)
            return False

    def doc_extension(self, prefere_extension: str = None) -> str:
        """
        It returns the extension of the task description file. If prefere_extension is not None,
        and the file with the given extension exists, it returns the given extension. Otherwise
        it returns the first found extension.

        :param prefere_extension: The preferred extension
        :type prefere_extension: str

        :return: The extension of the task description file
        :rtype: str

        :raise FileNotFoundError: If no file with valid extension exists.
        """
        if prefere_extension is not None:
            prefere_extension = self.DocExtension.from_str(prefere_extension)
            try:
                self.doc_path(prefere_extension)
                return prefere_extension.value
            except FileNotFoundError:
                pass

        for ext in self.DocExtension:
            try:
                self.doc_path(ext)
                return ext.value
            except FileNotFoundError:
                pass
        raise FileNotFoundError(f'No task description file found for {self.get("title")}')

    def get_docs_to_display(self, best_and_pdf: bool = True) -> List[Path]:
        """
        It returns the list of paths to the task description files.

        :return: The list of paths to the task description files
        :rtype: List[Path]
        """
        docs = []
        if not best_and_pdf:
            for ext in self.DocExtension:
                try:
                    docs.append(self.doc_path(ext))
                except FileNotFoundError:
                    pass
            return docs

        try:
            docs.append(self.doc_path('pdf'))
        except FileNotFoundError:
            pass
        for ext in self.DocExtension:
            if ext.value == 'pdf':
                continue
            try:
                docs.append(self.doc_path(ext))
                return docs
            except FileNotFoundError:
                pass

    def print_tree(self):
        """
        It prints the tree of the package
        """
        print(f'Package {self.name}:')

        for i in self._sets:
            print(f'|- {i["name"]}')
            for j in i._tests:
                print(f'|  |- {j["name"]}')
            print('|')


class TSet(PackageManager):
    """
    It's a class that represents a set of tests and modifies it
    """
    SETTINGS_VALIDATION = {
        'name': [[isStr]],
        'weight': [[isInt], [isFloat]],
        'points': [[isInt], [isFloat]],
        'memory_limit': [[isNone], [isSize, Package.MAX_SUBMIT_MEMORY]],
        'time_limit': [[isNone], [isIntBetween, 0, Package.MAX_SUBMIT_TIME],
                       [isFloatBetween, 0, Package.MAX_SUBMIT_TIME]],
        'checker': [[isNone], [isPath]],
        'verifier': [[isNone], [isPath]],
        'basename': [[isNone], [isPath]],
        'test_generator': [[isNone], [isPath]],
        'tests': [[isNone], [isAny]],
        'makefile': [[isNone], [isPath]]
    }

    """
        Validation for ``TSet`` settings.

        Available options are:
            * ``name``: set name
            * ``weight``: impact of set score on final score
            * ``points``: maximum amount of points to earn in set
            * ``memory_limit``: is a memory limit for set
            * ``time_limit``: is a time limit for set
            * ``checker``: is a path or None value to actual checker
            * ``test_generator``: is a path or None value to actual generator
            * ``tests``: tests to run in set
            * ``makefile``: name of a makefile
        """
    #: Default values for set settings
    DEFAULT_SETTINGS = {
        'name': '_unnamed',
        'weight': 1,
        'points': 0,
        'memory_limit': '512M',
        'time_limit': 1,
        'tests': {},
        'makefile': None,
        'checker': None,
        'verifier': None,
        'basename': None,
    }

    def __init__(self, path: Path, inherit_settings: dict = None):
        """
        It reads the config file and creates a list of tests

        :param path: Path - path to the test set
        :type path: Path
        """
        config_path = path / 'config.yml'
        if not config_path.is_file():
            config_path = path / 'config.yaml'
        if not config_path.is_file():
            config_path = {}
        super().__init__(path, config_path, TSet.DEFAULT_SETTINGS)
        if self['name'] == '_unnamed':
            self['name'] = path.name
        self._tests = []
        self._test_settings = {
            'name': '0',
            'memory_limit': self._settings['memory_limit'],
            'time_limit': self._settings['time_limit'],
            'points': 0
        }
        self.hinter = False

        if self._settings['tests'] is not None:
            for i in self._settings['tests'].values():
                self._tests.append(TestF(path, i, self._test_settings))
        if inherit_settings is not None:
            inherit_settings = inherit_settings.copy()
            if 'hinter' in inherit_settings.keys() and inherit_settings['hinter']:
                self.hinter = True
            for k, v in inherit_settings.items():
                if (k not in self._settings.keys() or not self._settings[k]) and v and k in self.DEFAULT_SETTINGS.keys():
                    self._settings[k] = v

        self._add_test_from_dir()

    def move_test_file(self, to_set, filename):
        """
        It moves a file from one directory to another

        :param to_set: the set you want to move the file to
        :param filename: the name of the file to move
        """
        if isPath(to_set._path):
            if isPath(self._path / filename):
                replace(self._path / filename, to_set._path / filename)

    def _add_test_from_dir(self):
        """
        It takes a directory, finds all the files in it, and then adds all the tests that have both an input and output file
        """
        test_files_ext = listdir(self._path)
        tests = []
        for i in test_files_ext:
            tests.append(match('.*[^.in|out]', i).group(0))
        tests_to_do = []
        for i in tests:
            if (not self.hinter and tests.count(i) == 2) or (self.hinter and tests.count(i) >= 1):
                tests_to_do.append(i)
        tests_to_do = set(tests_to_do)
        names = [i["name"] for i in self._tests]
        for i in tests_to_do:
            if i not in names:
                name_dict = {'name': i}
                self._tests.append(TestF(self._path, name_dict, self._test_settings))

    def tests(self, test_name: str = None, add_new: bool = False) -> 'TestF' or List['TestF']:
        """
        It returns a test object with the given name, if it exists, or creates a new one if it doesn't
        If no name is given, it returns a list of all the tests.

        :param test_name: The name of the test, if None, it returns a list of all the tests, defaults to None
        :type test_name: str
        :param add_new: if True, then if the test is not found, it will be created, defaults to False
        :type add_new: bool (optional)

        :return: A TestF object or a list of TestF objects
        :rtype: TestF or List[TestF]
        """
        if test_name is None:
            return self._tests
        for i in self._tests:
            if i['name'] == test_name:
                return i
        if add_new:
            new_test = TestF(self._path, {'name': test_name}, self._test_settings)
            self._tests.append(new_test)
            self.add_empty_file(test_name + '.in')
            self.add_empty_file(test_name + '.out')
            return new_test
        raise NoTestFound(f'Any test named {test_name} has found')

    def remove_file(self, filename):
        """
        It removes a file from the current directory

        :param filename: The name of the file to remove
        """
        if isPath(self._path / filename):
            remove(self._path / filename)

    def _delete_chosen_test(self, test: 'TestF'):
        """
        It removes the test from the list of tests, deletes the files associated with the test,
        and removes the test from the config file

        :param test: the test to be deleted
        :type test: 'TestF'
        """
        self._tests.remove(test)
        self.remove_file(test['name'] + '.in')
        self.remove_file(test['name'] + '.out')

        # removes settings for this test (from _settings and from config file)
        new_settings = deepcopy(self._settings)
        for k, v in self._settings['tests'].items():
            if v['name'] == test['name']:
                new_settings['tests'].pop(k)
        self._settings = new_settings
        self.save_to_config(self._settings)

    def delete_test(self, test_name: str):
        """
        It deletes a test from the list of tests, and deletes associated files.

        :param test_name: The name of the test to delete
        :type test_name: str
        :raise NoTestsFound: If there is no test with name equal to the test_name argument.
        """
        for test in self._tests:
            if test['name'] == test_name:
                self._delete_chosen_test(test)
                return

        raise NoTestFound(f'No test named {test_name} has found to delete')

    def _move_chosen_test(self, test: 'TestF', to_set: 'TSet'):
        """
         Move a test from one test set to another

        :param test: 'TestF' - the test to be moved
        :type test: 'TestF'
        :param to_set: the set to which the test will be moved
        :type to_set: 'TSet'
        """
        name_list_ta = [j['name'] for j in to_set._tests]
        if test['name'] not in name_list_ta:
            to_set._tests.append(test)
            self._tests.remove(test)
        else:
            raise TestExistError(f'Test named {test["name"]} exist in to_set files')

        self.move_test_file(to_set, test['name'] + '.in')
        self.move_test_file(to_set, test['name'] + '.out')

    def _move_config(self, to_set: 'TSet', test_name: str):
        """
        It takes a test name and a set object, and moves the test from the current set to the set object

        :param to_set: The set to move the test to
        :type to_set: 'TSet'
        :param test_name: The name of the test you want to move
        :type test_name: str
        """
        new_settings = deepcopy(self._settings)
        for i, j in self._settings['tests'].items():
            if j['name'] == test_name:
                to_set._settings['tests'][i] = j
                new_settings['tests'].pop(i)

        self._settings = new_settings
        self.remove_file('config.yml')
        self.save_to_config(self._settings)
        to_set.remove_file('config.yml')
        to_set.save_to_config(to_set._settings)

    # moves test to to_set (and all settings pinned to this test in self._settings)
    def move_test(self, test_name: str, to_set: 'TSet'):
        """
        It moves a test from one set to another

        :param test_name: The name of the test to move
        :type test_name: str
        :param to_set: The set to which the test will be moved
        :type to_set: 'TSet'
        """
        search = False
        for i in self._tests:
            if i['name'] == test_name:
                self._move_chosen_test(i, to_set)
                self._move_config(to_set, test_name)
                search |= True
                return
            search |= False

        if not search:
            raise NoTestFound(f'Any test named {test_name} has found to move to to_set')

    # check set validation
    def check_set(self, subtree=True) -> bool | int:
        """
        It checks the set.

        :param subtree: If True, check the subtree of tests, defaults to True (optional)

        :return: The result of the check_validation() method and the result of the check_set() method.
        """
        result = True
        if subtree:
            for i in self._tests:
                result &= i.check_test()

        return self.check_validation(TSet.SETTINGS_VALIDATION) & result


class TestF(PackageManager):
    """
    It's a class that represents test in a set.
    """
    SETTINGS_VALIDATION = {
        'name': [[isStr]],
        'memory_limit': [[isNone], [isSize, Package.MAX_SUBMIT_MEMORY]],
        'time_limit': [[isNone], [isIntBetween, 0, Package.MAX_SUBMIT_TIME],
                       [isFloatBetween, 0, Package.MAX_SUBMIT_TIME]],
        'points': [[isInt], [isFloat]],
        'input': [[isNone], [isPath]],
        'output': [[isNone], [isPath]],
    }
    """
       Validation for ``TestF`` settings.

       Available options are:
        * ``name``: test name
        * ``points``: maximum amount of points to earn in test
        * ``memory_limit``: is a memory limit for test
        * ``time_limit``: is a time limit for test
        * ``input``: is a path or None value to actual input
        * ``output``: is a path or None value to actual output
    """

    def __init__(self, path: Path, additional_settings: dict or Path, default_settings: dict, hinter_mode: bool = False):
        """
        This function initializes the class by calling the superclass's __init__ function, which is the __init__ function of
        the Config class

        :param path: The path to the file that contains the settings
        :type path: Path
        :param additional_settings: This is a dictionary of settings that you want to override the default settings with
        :type additional_settings: dict or Path
        :param default_settings: This is a dictionary of default settings that will be used if the settings file doesn't
        contain a value for a setting
        :type default_settings: dict
        """
        super().__init__(path, additional_settings, default_settings)
        input_file = self._path / (self._settings['name'] + '.in')
        if input_file.is_file():
            self._settings['input'] = input_file
        output_file = self._path / (self._settings['name'] + '.out')
        if not hinter_mode and output_file.is_file():
            self._settings['output'] = output_file

    def _rename_files(self, old_name, new_name):
        """
        It renames a file

        :param old_name: The name of the file you want to rename
        :param new_name: The new name of the file
        """
        rename(self._path / old_name, self._path / new_name)

    def __setitem__(self, arg: str, val):
        """
        It renames the files and changes the name in the yaml file

        :param arg: the key of the dictionary
        :type arg: str
        :param val: the new value
        """
        # effect changes to yaml settings
        settings = self.read_from_config()
        if arg == 'name':
            self._rename_files(self._settings['name'] + '.in', val + '.in')
            self._rename_files(self._settings['name'] + '.out', val + '.out')

            for i, j in settings['tests'].items():
                if j['name'] == self._settings['name']:
                    new_key = 'test' + val
                    j[arg] = val
                    settings['tests'][new_key] = j
                    del settings['tests'][i]
                    break

        self.save_to_config(settings)
        self._settings[arg] = val

    def check_test(self) -> bool:
        """
        It checks if the test is valid
        :return: The return value is the result of the check_validation method.
        """
        return self.check_validation(TestF.SETTINGS_VALIDATION)



import shutil
from pathlib import Path
from unittest import TestCase

from baca2PackageManager.zipper import Zip


class UnzipTest(TestCase):
    def setUp(self) -> None:
        self.path = Path('./test_packages')
        if not self.path.is_dir():
            self.path = Path('./tests/test_packages')
        self.src = self.path / 'zip_src'
        self.dist = self.path / 'zip_dst'
        shutil.rmtree(self.dist, ignore_errors=True)

    def tearDown(self) -> None:
        shutil.rmtree(self.dist, ignore_errors=True)

    def test_01_simple_unzip(self):
        with Zip(self.src / 'zok.zip') as zip_f:
            zip_f.extractall(self.dist)
        dir_path = Path(self.dist / 'zok')
        self.assertTrue(dir_path.is_dir())
        arr = [f for f in dir_path.iterdir()]
        self.assertEqual(len(arr), 3)
        with open(dir_path / '1.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/1.txt ok')
        with open(dir_path / '2.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/2.txt ok')
        with open(dir_path / '3.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/3.txt ok')

    def test_02_zipbomb_detection(self):
        with Zip(self.src / 'zbsm.zip') as zip_f:
            with self.assertRaises(Zip.UnsafeZipFile):
                zip_f.extractall(self.dist)

    def test_03_nested_zipfile_extraction(self):
        with Zip(self.src / 'znest.zip') as zip_f:
            zip_f.extractall(self.dist)
        dir_path = Path(self.dist / 'znest')
        self.assertTrue(dir_path.is_dir())
        arr = [f for f in dir_path.iterdir()]
        self.assertEqual(len(arr), 3)
        with open(dir_path / '1.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/1.txt ok')
        with open(dir_path / '2.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/2.txt ok')
        with open(dir_path / '3.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/3.txt ok')

    def test_04_nested_zipfile_without_leave_top(self):
        with Zip(self.src / 'znest.zip') as zip_f:
            zip_f.extractall(self.dist, leave_top=False)
        dir_path = Path(self.dist)
        self.assertTrue(dir_path.is_dir())
        arr = [f for f in dir_path.iterdir()]
        self.assertEqual(len(arr), 3)
        with open(dir_path / '1.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/1.txt ok')
        with open(dir_path / '2.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/2.txt ok')
        with open(dir_path / '3.txt', 'r') as f:
            self.assertEqual(f.read(), 'zok/3.txt ok')

    def test_05_source_extensions_allowed(self):
        with Zip(self.src / 'z_src_cpp.zip') as zip_f:
            self.assertTrue(zip_f.check_extensions(['cpp']))
            self.assertFalse(zip_f.check_extensions(['c', 'h', 'hpp']))

    def test_06_source_extensions_not_allowed(self):
        with Zip(self.src / 'z_src_cpp_py.zip') as zip_f:
            self.assertFalse(zip_f.check_extensions(['cpp']))
            self.assertTrue(zip_f.check_extensions(['cpp', 'py']))



import unittest as ut

import os
from copy import deepcopy
from typing import Any

import baca2PackageManager.judge_manager as judge


class JNAlpha(judge.JudgeNodeBase):

    def start(self, *args, **kwargs) -> Any:
        return None

    def receive(self, *args, **kwargs) -> judge.JudgeVerdict:
        assert len(args) > 0
        return judge.JudgeVerdict.OK if args[0] else judge.JudgeVerdict.FAIL


class JNBeta(judge.JudgeNodeBase):

    def start(self, *args, **kwargs) -> Any:
        return None

    def receive(self, *args, **kwargs) -> judge.JudgeVerdict:
        assert len(args) > 0
        return judge.JudgeVerdict.FAIL if args[0] else judge.JudgeVerdict.OK


class JNTheta(judge.JudgeNodeBase):

    def start(self, *args, **kwargs) -> Any:
        return None

    def receive(self, *args, **kwargs) -> judge.JudgeVerdict:
        return judge.JudgeVerdict.OK


class GeneralTest(ut.TestCase):

    def setUp(self) -> None:
        self.a = JNAlpha('a')
        self.a2 = JNAlpha('a2')
        self.b = JNBeta('b')
        self.c = JNTheta('c')
        self.end = judge.EndNode('END', judge.JudgeVerdict.OK)

    def tearDown(self) -> None:
        files_to_delete = ['serialization_test.bin',
                           'serialization_graph_test.bin',
                           'serialization_complex_graph_test.bin',
                           'serialization_test.yaml',
                           'serialization_graph_test.yaml',
                           'serialization_complex_graph_test.yaml',
                           ]
        for file in files_to_delete:
            try:
                os.remove(file)
            except FileNotFoundError:
                pass

    def _from_dict_func(self) -> tuple[dict, judge.JudgeManager]:
        default = {
            self.a: {judge.JudgeVerdict.OK: self.a2, judge.JudgeVerdict.FAIL: self.a},
            self.a2: {judge.JudgeVerdict.OK: self.b},
            self.b: {judge.JudgeVerdict.OK: self.c, judge.JudgeVerdict.FAIL: self.a},
            self.c: {judge.JudgeVerdict.OK: self.end, judge.JudgeVerdict.FAIL: self.b},
            self.end: {}
        }
        return default, judge.JudgeManager.from_dict(default, self.a)

    def test_from_dict(self):
        default, manager = self._from_dict_func()
        self.assertEqual(manager.graph, default)

    def test_adding(self):
        manager = judge.JudgeManager()
        manager.add_node(self.a)
        self.assertRaises(ValueError, manager.add_node, self.a)
        self.assertRaises(KeyError, manager.add_connection, self.a, self.b, judge.JudgeVerdict.OK)
        self.assertRaises(KeyError, manager.add_connection, self.b, self.a, judge.JudgeVerdict.OK)
        manager.add_node(self.b)
        self.assertTrue(self.a in manager.graph and manager.graph[self.a] == {} and
                        self.b in manager.graph and manager.graph[self.b] == {})
        manager.add_connection(self.a, self.b, judge.JudgeVerdict.OK)
        self.assertEqual(manager.graph[self.a], {judge.JudgeVerdict.OK: self.b})
        manager.add_node(self.c)
        manager.add_connection(self.a, self.c, judge.JudgeVerdict.FAIL)
        self.assertEqual(manager.graph[self.a], {judge.JudgeVerdict.OK: self.b, judge.JudgeVerdict.FAIL: self.c})
        manager.add_node(self.end)
        self.assertRaises(ResourceWarning, manager.add_connection, self.end, self.a, judge.JudgeVerdict.OK)

    def test_removing(self):
        manager = judge.JudgeManager()
        manager.add_node(self.a)
        self.assertRaises(KeyError, manager.remove_node, self.b)
        manager.add_node(self.b)
        manager.add_node(self.c)
        self.assertRaises(KeyError, manager.remove_connection_by_node, self.a, self.b)
        self.assertRaises(KeyError, manager.remove_connection_by_verdict, self.a, judge.JudgeVerdict.FAIL)
        manager.add_connection(self.a, self.b, judge.JudgeVerdict.OK)
        manager.add_connection(self.a, self.c, judge.JudgeVerdict.FAIL)
        manager.add_connection(self.a, self.c, judge.JudgeVerdict.INCONCLUSIVE)
        manager.remove_node(self.b)
        self.assertTrue(manager.graph[self.a] == {judge.JudgeVerdict.FAIL: self.c,
                                                  judge.JudgeVerdict.INCONCLUSIVE: self.c})
        manager.remove_connection_by_node(self.a, self.c)
        self.assertTrue(manager.graph[self.a] == {} and self.c in manager.graph)
        manager.add_connection(self.a, self.c, judge.JudgeVerdict.FAIL)
        manager.remove_connection_by_verdict(self.a, judge.JudgeVerdict.FAIL)
        self.assertTrue(manager.graph[self.a] == {} and self.c in manager.graph)

    def test_send_receive(self):
        manager = judge.JudgeManager()
        manager.add_node(self.a)
        manager.add_node(self.b)
        manager.add_node(self.end)
        manager.add_connection(self.a, self.b, judge.JudgeVerdict.OK)
        manager.add_connection(self.b, self.end, judge.JudgeVerdict.OK)
        manager.add_connection(self.b, self.a, judge.JudgeVerdict.FAIL)
        self.assertEqual(manager.send(self.a, True), None)
        self.assertRaises(KeyError, manager.send, self.a2)
        self.assertRaises(judge.EndNodeError, manager.advance, self.end, judge.JudgeVerdict.OK)
        manager.set_start_node(self.a)
        current = manager.get_start_node()
        current = manager.advance(current, judge.JudgeVerdict.OK)
        self.assertEqual(current, self.b)
        current = manager.advance(current, judge.JudgeVerdict.FAIL)
        self.assertEqual(current, self.a)
        current = manager.advance(current, judge.JudgeVerdict.OK)
        current = manager.advance(current, judge.JudgeVerdict.OK)
        self.assertEqual(current, self.end)
        self.assertRaises(judge.EndNodeError, manager.advance, current, judge.JudgeVerdict.OK)

    def test_check_integrity_simple(self):
        _, manager = self._from_dict_func()
        out = manager._floyd_warshall()
        expected = {
            self.a: frozenset({self.a, self.a2, self.b, self.c, self.end}),
            self.a2: frozenset({self.a, self.a2, self.b, self.c, self.end}),
            self.b: frozenset({self.a, self.a2, self.b, self.c, self.end}),
            self.c: frozenset({self.a, self.a2, self.b, self.c, self.end}),
            self.end: frozenset({self.end})
        }
        self.assertEqual(out, expected)

        report = manager.check_graph_integrity()
        self.assertEqual(frozenset(report.wrong_connections), frozenset({self.a2}))
        self.assertEqual(report.cannot_reach_end, [])
        self.assertEqual(report.unreachable_nodes, [])
        self.assertEqual(report.has_end_nodes, True)
        self.assertEqual(report.no_errors, False)

        manager.add_connection(self.a2, self.a, judge.JudgeVerdict.FAIL)
        report = manager.check_graph_integrity()
        self.assertTrue(report.no_errors)

    def test_check_integrity_complex(self):
        tmp = {
            self.a: {judge.JudgeVerdict.INCONCLUSIVE: self.a, judge.JudgeVerdict.OK: self.b},
            self.a2: {judge.JudgeVerdict.FAIL: self.a2, judge.JudgeVerdict.OK: self.b},
            self.b: {judge.JudgeVerdict.INCONCLUSIVE: self.c, judge.JudgeVerdict.FAIL: self.end},
            self.c: {judge.JudgeVerdict.OK: self.c},
            self.end: {}
        }
        manager = judge.JudgeManager.from_dict(tmp, self.a)

        out = manager._floyd_warshall()
        expected = {
            self.a: {self.a, self.b, self.c, self.end},
            self.a2: {self.a2, self.b, self.c, self.end},
            self.b: {self.b, self.c, self.end},
            self.c: {self.c},
            self.end: {self.end}
        }
        self.assertEqual(out, expected)

        report = manager.check_graph_integrity()
        self.assertEqual(report.has_end_nodes, True)
        self.assertEqual(frozenset(report.unreachable_nodes), frozenset({self.a2}))
        self.assertEqual(frozenset(report.cannot_reach_end), frozenset({self.c}))
        self.assertEqual(frozenset(report.wrong_connections), frozenset({self.b, self.c}))

        manager.remove_node(self.end)
        report = manager.check_graph_integrity()
        self.assertEqual(report.has_end_nodes, False)
        self.assertEqual(frozenset(report.unreachable_nodes), frozenset({self.a2}))
        self.assertEqual(frozenset(report.cannot_reach_end), frozenset({self.a, self.b, self.a2, self.c}))
        self.assertEqual(frozenset(report.wrong_connections), frozenset({self.b, self.c}))

    def test_node_yaml_serialisation(self):
        q = deepcopy(self.a).serialise()
        q = judge.JudgeNodeBase.unpack(q)
        self.assertEqual(vars(self.a), vars(q))

    def test_node_yaml_serialisation_to_file(self):
        with open('serialization_test.bin', 'wb') as f:
            q = deepcopy(self.a).serialise()
            f.write(q)

        with open('serialization_test.bin', 'rb') as f:
            q = f.read()
            q = judge.JudgeNodeBase.unpack(q)
        self.assertEqual(vars(self.a), vars(q))

    def test_graph_yaml_serialisation_to_file(self):
        graph = judge.JudgeManager.from_dict({self.a: {judge.JudgeVerdict.OK: self.a}}, self.a)

        with open('serialization_graph_test.bin', 'wb') as f:
            q = deepcopy(graph).serialise()
            f.write(q)

        with open('serialization_graph_test.bin', 'rb') as f:
            q = f.read()
            q = judge.JudgeManager.unpack(q)
        self.assertEqual(len(q.graph), 1)
        self.assertEqual(q.graph[self.a][judge.JudgeVerdict.OK], self.a)

    def test_complex_graph_yaml_serialisation_to_file(self):
        _, graph = self._from_dict_func()

        with open('serialization_complex_graph_test.bin', 'wb') as f:
            q = deepcopy(graph).serialise()
            f.write(q)

        with open('serialization_complex_graph_test.bin', 'rb') as f:
            q = f.read()
            q = judge.JudgeManager.unpack(q)

        self.assertEqual(q.graph, graph.graph)



import shutil
from unittest import TestCase

from parameterized import parameterized
import pytest

from baca2PackageManager.manager_exceptions import NoSetFound, NoTestFound, InvalidFileExtension
from baca2PackageManager.validators import isAny, isNone, isInt, isIntBetween, isFloat, \
    isFloatBetween, isStr, is_, \
    isIn, isShorter, isDict, isPath, isSize, isList, valid_memory_size, \
    hasStructure, isSize
from baca2PackageManager.tools import bytes_from_str
from baca2PackageManager import TestF as TF, TSet, Package, set_base_dir, base_dir, \
    add_supported_extensions
from pathlib import Path
from yaml import safe_load
import random
import string

add_supported_extensions('cpp')
set_base_dir(Path(__file__).parent / 'test_packages')


def generate_rand_int():
    return random.randint(0, 10000000)


def compare_memory_unit(unit1, unit2):
    units = ['B', 'K', 'M', 'G']
    try:
        index1 = units.index(unit1)
        index2 = units.index(unit2)
        return index1 <= index2
    except ValueError:
        return False


def generate_rand_dict():
    _dict = {}
    for i in range(100):
        key1 = random.choice(string.ascii_letters)
        value1 = random.randint(1, 10000)
        key2 = random.randint(1, 10000)
        value2 = random.choice(string.ascii_letters)
        if i % 4 == 0:
            _dict[key1] = value1
        elif i % 4 == 1:
            _dict[key2] = value2
        elif i % 4 == 2:
            _dict[key1] = value2
        else:
            _dict[key2] = value1
    return _dict


def generate_rand_list(only_int=False, only_str=False, only_float=False):
    _list = []
    for i in range(100):
        if only_int:
            val = random.randint(1, 10000)
        elif only_str:
            val = random.choice(string.ascii_letters)
        elif only_float:
            val = random.uniform(1, 10000)
        else:
            val = random.choice([random.randint(1, 10000),
                                 random.choice(string.ascii_letters),
                                 random.uniform(1, 10000)])
        _list.append(val)
    return _list


class ValidationsTests(TestCase):

    def test_isInt(self):
        """
        It tests if the function isInt() works correctly
        """
        for i in range(1000):
            a = generate_rand_int()
            self.assertTrue(isInt(a))
            float(a)
            self.assertTrue(isInt(a))
            str(a)
            self.assertTrue(isInt(a))

        self.assertFalse(isInt(0.5))
        self.assertFalse(isInt('0.5'))
        self.assertFalse(isInt(1235.4567))
        self.assertFalse(isInt('1235.4567'))
        self.assertFalse(isInt('5,5'))
        self.assertFalse(isInt("5 and more"))
        self.assertFalse(isInt("It is just a string"))

    def test_isIntBetween(self):  # a <= val < b
        """
        It tests if the function is_IntBetween() works correctly        """
        for i in range(1000):
            a = generate_rand_int()
            b = generate_rand_int()
            if a > b:
                a, b = b, a
            val = random.randint(a, b - 1)
            self.assertTrue(isIntBetween(val, a, b))
            float(a)
            float(b)
            float(val)
            self.assertTrue(isIntBetween(val, a, b))
            val = random.randint(0, a - 1)
            self.assertFalse(isIntBetween(val, a, b))
            val = random.randint(b + 1, 10000001)
            self.assertFalse(isIntBetween(val, a, b))
        self.assertTrue(isIntBetween(5, 2, 10))
        self.assertTrue(isIntBetween(68, 68, 78))
        self.assertTrue(isIntBetween(-4, -7, -1))
        self.assertTrue(isIntBetween(-6, -6, 0))
        self.assertFalse(isIntBetween(5, 1, 3))
        self.assertFalse(isIntBetween(5, -7, -1))
        self.assertFalse(isIntBetween(5, -6, 0))
        self.assertFalse(isIntBetween(5, 6, 67))
        self.assertFalse(isIntBetween(67, 6, 67))

    def test_isFloat(self):
        """
        It tests if the function isFloat() works correctly
        """
        for i in range(1000):
            a = random.uniform(0.1, 100000.0)
            self.assertTrue(isFloat(a))
            float(a)
            self.assertTrue(isFloat(a))
            str(a)
            self.assertTrue(isFloat(a))
            a = random.random()
            self.assertTrue(isFloat(a))
            float(a)
            self.assertTrue(isFloat(a))
            str(a)
            self.assertTrue(isFloat(a))
        self.assertTrue(isFloat('5'))
        self.assertTrue(isFloat(5))
        self.assertTrue(isFloat(0))
        self.assertTrue(isFloat('0'))
        self.assertTrue(isFloat(123456))
        self.assertTrue(isFloat('123456'))
        self.assertTrue(isFloat(0.5))
        self.assertTrue(isFloat('0.5'))
        self.assertTrue(isFloat(1235.4567))
        self.assertTrue(isFloat('1235.4567'))
        self.assertFalse(isFloat('5,5'))
        self.assertFalse(isFloat("5 and more"))
        self.assertFalse(isFloat("53.46 and more"))
        self.assertFalse(isFloat("It is just a string"))

    def test_isFloatBetween(self):
        """
        It tests if the function isFloatBetween() works correctly
        """
        for i in range(1000):
            a = random.randint(1, 10000000)
            b = random.randint(1, 10000000)
            if a > b:
                a, b = b, a
            val = random.uniform(a, b - 1)
            self.assertTrue(isFloatBetween(val, a, b))
            float(a)
            float(b)
            float(val)
            self.assertTrue(isFloatBetween(val, a, b))
            val = random.uniform(0.0, a - 1.0)
            self.assertFalse(isFloatBetween(val, a, b))
            val = random.uniform(b + 1.0, 10000001.0)
            self.assertFalse(isFloatBetween(val, a, b))

    def test_isStr(self):
        """
        It tests if the function isStr() works correctly
        """
        for i in range(1000):
            val = random.choice(string.ascii_letters)
            self.assertTrue(isStr(val))
        for i in range(10000):
            val2 = random.randint(1, 10000000)
            self.assertFalse(isStr(val2))
            float(val2)
            self.assertFalse(isStr(val2))

    def test_is_(self):
        """
        It tests if the function is_() works correctly
        """
        for i in range(1000):
            val = random.choice(string.ascii_letters)
            schema = val
            self.assertTrue(is_(val, schema))
            schema = random.choice(string.ascii_letters)
            if (schema != val):
                self.assertFalse(is_(val, schema))
            val = random.randint(1, 100000)
            self.assertFalse(is_(val, schema))
            float(val)
            self.assertFalse(is_(val, schema))

    def test_isShorter(self):
        """
        It tests if the function isShorter() works correctly
        """
        for i in range(1000):
            val = random.choice(string.ascii_letters)
            _int = random.randint(len(val) + 1, 10000)
            self.assertTrue(isShorter(val, _int))
            _int = random.randint(0, len(val) - 1)
            self.assertFalse(isShorter(val, _int))
        for i in range(10000):
            val2 = random.randint(1, 10000000)
            _int = random.randint(1, 10000000)
            self.assertFalse(isShorter(val2, _int))
            float(val2)
            self.assertFalse(isShorter(val2, _int))

    def test_isDict(self):
        """
        It tests if the function isDict() works correctly
        """
        for i in range(1000):
            _dict = generate_rand_dict()
            self.assertTrue(isDict(_dict))
        for i in range(1000):
            value = random.randint(1, 10000)
            self.assertFalse(isDict(value))
            float(value)
            self.assertFalse(isDict(value))
            value = random.choice(string.ascii_letters)
            self.assertFalse(isDict(value))

    def test_isList(self):
        """
        It tests if the function isList() works correctly
        """
        for i in range(1000):
            _list = generate_rand_list()
            self.assertTrue(isList(_list))

    def test_isList_with_validation(self):
        """
        It tests if the function isList() works correctly
        """
        for i in range(1000):
            _list = generate_rand_list(only_int=True)
            self.assertTrue(isList(_list, [isInt]))
            self.assertTrue(isList(_list, [isFloat]))
            self.assertFalse(isList(_list, [isStr]))
            self.assertTrue(isList(_list, [isAny]))
            self.assertTrue(isList(_list, [isIntBetween, 0, 10001]))
            self.assertFalse(isList(_list, [isDict]))
            self.assertFalse(isList(_list, [isList]))
            self.assertFalse(isList(_list, [isNone]))
        for i in range(1000):
            _list = generate_rand_list(only_str=True)
            self.assertFalse(isList(_list, [isInt]))
            self.assertFalse(isList(_list, [isFloat]))
            self.assertTrue(isList(_list, [isStr]))
            self.assertTrue(isList(_list, [isAny]))
            self.assertFalse(isList(_list, [isIntBetween, 0, 10001]))
            self.assertFalse(isList(_list, [isDict]))
            self.assertFalse(isList(_list, [isList]))
            self.assertFalse(isList(_list, [isNone]))
        for i in range(1000):
            _list = generate_rand_list(only_float=True)
            self.assertFalse(isList(_list, [isInt]))
            self.assertTrue(isList(_list, [isFloat]))
            self.assertFalse(isList(_list, [isStr]))
            self.assertTrue(isList(_list, [isAny]))
            self.assertTrue(isList(_list, [isFloatBetween, 0, 10000]))
            self.assertFalse(isList(_list, [isDict]))
            self.assertFalse(isList(_list, [isList]))
            self.assertFalse(isList(_list, [isNone]))

    def test_memory_converting(self):
        """
        It tests if the function memory_converting() works correctly
        """
        # error message in case if test case got failed
        message = "First value and second value are not equal!"
        # assertEqual() to check equality of first & second value
        size = 456
        val = str(size) + "B"
        # test for B unit value
        self.assertEqual(bytes_from_str(val), size, message)
        val = str(size) + "K"
        # test for K unit value
        self.assertEqual(bytes_from_str(val), size * 1024, message)
        val = str(size) + "M"
        # test for M unit value
        self.assertEqual(bytes_from_str(val), size * 1024 * 1024, message)
        val = str(size) + "G"
        # test for G unit value
        self.assertEqual(bytes_from_str(val), size * 1024 * 1024 * 1024, message)

    def test_valid_memory_size(self):
        """
        It tests if the function memory_size() works correctly
        """
        unit_list = ['B', 'K', 'M', 'G']
        for i in range(1000):
            size = random.randint(1, 100000)
            max_size = random.randint(size, 100001)
            unit1 = random.choice(unit_list)
            unit2 = random.choice(unit_list)
            if not (compare_memory_unit(unit1, unit2)):
                unit1, unit2 = unit2, unit1  # now unit1 is smaller than unit
            mem_size = str(size) + unit1
            max_mem_size = str(max_size) + unit2
            self.assertTrue(valid_memory_size(mem_size, max_mem_size))

    def test_hasStructure(self):
        """
        It tests if the function hasStructure() works correctly
        """
        # validator at the end of a string
        structure = "set<isInt>"
        self.assertTrue(hasStructure("set123", structure))
        self.assertTrue(hasStructure("set", structure))
        structure = "test_<isIn,'a','wrong','0'>"
        self.assertTrue(hasStructure("test_a", structure))
        self.assertTrue(hasStructure("test_wrong", structure))
        self.assertTrue(hasStructure("test_0", structure))
        # structure at the end and in the middle
        structure = "test<isInt>_set<isInt>"
        self.assertTrue(hasStructure("test123_set123", structure))
        self.assertFalse(hasStructure("test_13_set12", structure))
        self.assertFalse(hasStructure("test12set34", structure))
        self.assertFalse(hasStructure("test123_set", structure))
        self.assertFalse(hasStructure("test_123_set", structure))
        self.assertFalse(hasStructure("test_123", structure))
        # structure at the beginning
        structure = "<isInt>.in"
        self.assertFalse(hasStructure("1. in", structure))
        self.assertFalse(hasStructure("str.in", structure))
        self.assertFalse(hasStructure("2.5.in", structure))

        # two structures with alternative at beginning
        structure = "<isStr>|<isInt>_course"
        self.assertTrue(hasStructure("ASD_course", structure))
        self.assertTrue(hasStructure("MD_course", structure))
        self.assertTrue(hasStructure("123_course", structure))
        # structure at the beginning, middle and at the end
        structure = "<isInt>test<isStr>|<isInt>|<isFloat>"
        self.assertTrue(hasStructure("23test23", structure))
        self.assertTrue(hasStructure("123testSTR", structure))
        self.assertTrue(hasStructure("123test5.67", structure))
        self.assertFalse(hasStructure("test", structure))
        self.assertFalse(hasStructure("34TEST23", structure))

    def test_isAny(self):
        """
        It tests if the function isAny() works correctly
        """
        for i in range(1000):
            val = random.randint(0, 10000000)
            self.assertTrue(isAny(val))
            float(val)
            self.assertTrue(isAny(val))
            str(val)
            self.assertTrue(isAny(val))
            val = random.uniform(0.1, 100000.0)
            self.assertTrue(isAny(val))
            float(val)
            self.assertTrue(isAny(val))
            str(val)
            self.assertTrue(isAny(val))
            a = random.random()
            self.assertTrue(isAny(val))
            float(val)
            self.assertTrue(isAny(val))
            str(val)
            self.assertTrue(isAny(val))
            val = random.choice(string.ascii_letters)
            self.assertTrue(isAny(val))
            _dict = generate_rand_dict()
            self.assertTrue(isAny(_dict))
            _list = generate_rand_list()
            self.assertTrue(isAny(_list))

    def test_isNone(self):
        """
        It tests if the function isNone() works correctly
        """
        for i in range(1000):
            val = random.randint(0, 10000000)
            self.assertFalse(isNone(val))
            float(val)
            self.assertFalse(isNone(val))
            str(val)
            val = random.choice(string.ascii_letters)
            self.assertFalse(isNone(val))
            _dict = generate_rand_dict()
            self.assertFalse(isNone(_dict))
            _list = generate_rand_list()
            self.assertFalse(isNone(_list))
            self.assertTrue(isNone(None))

    def test_isIn(self):
        """
        It tests if the function isIn() works correctly
        """
        rand_int = generate_rand_int()
        rand_float = random.uniform(0.1, 0.9)
        rand_str = random.choice(string.ascii_letters)
        rand_dict = generate_rand_dict()
        rand_list = generate_rand_list()
        self.assertTrue(isIn(rand_int, rand_int, rand_float, rand_str, rand_dict, rand_list))
        self.assertTrue(isIn(rand_float, rand_int, rand_float, rand_str, rand_dict, rand_list))
        self.assertTrue(isIn(rand_str, rand_int, rand_float, rand_str, rand_dict, rand_list))
        self.assertTrue(isIn(rand_dict, rand_int, rand_float, rand_str, rand_dict, rand_list))
        self.assertTrue(isIn(rand_list, rand_int, rand_float, rand_str, rand_dict, rand_list))
        self.assertFalse(isIn(rand_int, rand_float, rand_str, rand_dict))
        self.assertFalse(isIn(rand_list, rand_float, rand_str, rand_dict))
        self.assertFalse(isIn(rand_dict, rand_float, rand_str, rand_list))

    # def test_isPath(self):
    #     """
    #     It tests if the function isPath() works correctly.
    #     """
    #     abs_path = Path("BaCa2/package/packages/tests_to_testing/config.yml").resolve()
    #     self.assertTrue(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/tests_to_testing/1.out").resolve()
    #     self.assertTrue(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/1/config.yml").resolve()
    #     self.assertTrue(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/1/prog/solution.cpp").resolve()
    #     self.assertTrue(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/tests_to_testing/10.out").resolve()
    #     self.assertFalse(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/1/config234.yml").resolve()
    #     self.assertFalse(isPath(abs_path))
    #     abs_path = Path("BaCa2/package/packages/1/prog/sopution.cpp").resolve()
    #     self.assertFalse(isPath(abs_path))

    def size_test(self, size_min, size_max):
        pass

    @parameterized.expand([
        (1, 100),
        (100, 1000),
        (1000, 10000),
        (1, 100000)
    ])
    def test_isSize(self, size_min, size_max):
        """
        It tests if the function isSize() works correctly
        """
        unit_list = ['B', 'K', 'M', 'G']
        for i in range(3000):
            size = random.randint(size_min, size_max)
            max_size = random.randint(size, size_max + 1)
            unit1 = random.choice(unit_list)
            unit2 = random.choice(unit_list)
            if not (compare_memory_unit(unit1, unit2)):
                unit1, unit2 = unit2, unit1  # now unit1 is smaller than unit
            mem_size = str(size) + unit1
            max_mem_size = str(max_size) + unit2
            self.assertTrue(isSize(mem_size, max_mem_size),
                            f"mem_size: {mem_size}, max_mem_size: {max_mem_size}")


# It tests the TestF class
class TestFTests(TestCase):
    def setUp(self):
        """
        It creates a test object with the following parameters:
        
        * `Path(__file__).resolve().parent / 'packages' / 'tests_to_testing'` - the path to the directory containing the
        test files.
        * `{}` - the input data for the test.
        * `{'name': '1', 'memory_limit': '10M', 'time_limit': 5}` - the test parameters
        """
        self.test = TF(base_dir() / 'tests_to_testing', {}, {
            'name': '1',
            'memory_limit': '10M',
            'time_limit': 5
        })

    def test_check_valid(self):
        """
        It checks if the test is valid.
        """
        self.assertTrue(self.test.check_test())

    def test_get_settings(self):
        """
        This function tests the get_settings function in the test_settings.py file
        """
        self.assertEqual(self.test['name'], '1')
        self.assertEqual(self.test['memory_limit'], '10M')
        self.assertEqual(self.test['time_limit'], 5)

    def test_set_settings(self):
        """
        It tests that the settings can be set
        """
        self.test['name'] = '2'
        self.assertEqual(self.test['name'], '2')
        self.test['name'] = '1'


# `TestTSet` is a class that tests the class `TSet`
class TSetTests(TestCase):
    def setUp(self):
        """
        The function `setUp` is a method of the class `TestTSet` that creates three instances of the class `TSet` and
        assigns them to the variables `self.set0`, `self.set1`, and `self.set2`
        """
        self.path = base_dir() / '1' / '1' / 'tests'
        self.set0 = TSet(self.path / 'set0')
        self.set1 = TSet(self.path / 'set1')
        self.set2 = TSet(self.path / 'set2')

    def test_init(self):
        """
        `test_init` checks that the `_path` attribute of the `Set` object is equal to the path of the set
        """
        self.assertEqual(self.set0._path, self.path / 'set0')
        self.assertEqual(self.set1._path, self.path / 'set1')
        self.assertEqual(self.set2._path, self.path / 'set2')

    def test_check_valid(self):
        """
        This function checks if the set is valid
        """
        self.assertTrue(self.set0.check_set())
        self.assertEqual(len(self.set0._tests), 2)
        self.assertTrue(self.set1.check_set())
        self.assertTrue(self.set2.check_set())

    def test_func_tests1(self):
        self.assertEqual(self.set0.tests('2'), self.set0._tests[0])
        test = [i for i in self.set1._tests if i['name'] == '2']
        self.assertEqual(self.set1.tests('2'), test[0])
        self.assertEqual(self.set2.tests('3', True), self.set2._tests[3])
        self.set2.delete_test('3')

    def test_func_tests2(self):
        """
        It tests the function tests() in the class TestSet.
        """
        self.assertRaises(NoTestFound, self.set0.tests, '1')
        self.assertRaises(NoTestFound, self.set1.tests, '3')
        self.assertRaises(NoTestFound, self.set2.tests, '2')

    def test_func_delete1(self):
        """
        This function tests the delete_test function in the TestSet class
        """
        self.assertEqual(self.set0.tests('4', True), self.set0._tests[2])
        self.set0.delete_test('4')
        names = [i['name'] for i in self.set0._tests]
        self.assertNotIn('4', names)

    def test_func_delete2(self):
        """
        It checks if the test is in the set.
        """
        self.assertRaises(NoTestFound, self.set0.tests, '1')
        self.assertRaises(NoTestFound, self.set1.tests, '3')
        self.assertRaises(NoTestFound, self.set2.tests, '2')

    def test_func_move1(self):
        """
        It moves a test from one test set to another.
        """
        self.set2.move_test('5', self.set1)
        names = [i['name'] for i in self.set1._tests]
        self.assertIn('5', names)
        self.set1.move_test('5', self.set2)

    def test_func_move2(self):
        """
        It moves a test from one set to another.
        """
        self.set2.move_test('6', self.set1)
        with open(self.set1._path / 'config.yml', mode="rt", encoding="utf-8") as file:
            settings = safe_load(file)
        value = False
        for i in settings['tests'].values():
            if i['name'] == '6':
                value = True
        self.assertTrue(value)
        self.set1.move_test('6', self.set2)

    def test_func_move2(self):
        """
        It tests if the function tests() raises an error when the test number is not found.
        """
        self.assertRaises(NoTestFound, self.set0.tests, '1')
        self.assertRaises(NoTestFound, self.set1.tests, '3')
        self.assertRaises(NoTestFound, self.set2.tests, '2')


# It tests the package class
class PackageTests(TestCase):
    def setUp(self):
        """
        The function takes a path to a package and returns a package object
        """
        self.path = base_dir() / '1'
        self.package = Package(self.path, commit='1')
        self.path = self.path / '1'

    def test_init(self):
        """
        The function tests that the package's path is the same as the path that was passed to the function, and that the
        package's title is the same as the title in the config.yml file
        """
        self.assertEqual(self.package.commit_path, self.path)
        with open(self.path / 'config.yml', mode="rt", encoding="utf-8") as file:
            settings = safe_load(file)
        self.assertEqual(self.package['title'], settings['title'])

    def test_func_sets1(self):
        """
        It tests if the function sets() returns the correct set.
        """
        self.assertIn(self.package.sets('set0'), self.package._sets)

    def test_func_sets2(self):
        """
        It tests that the function sets() raises an exception when it is passed a set that does not
        exist.
        """
        self.assertRaises(NoSetFound, self.package.sets, 'set3')

    def test_func_sets3(self):
        """
        This function tests the sets function in the package class
        """
        self.package.sets('set4', True)
        sets = [i['name'] for i in self.package._sets]
        self.assertIn('set4', sets)
        self.assertTrue(self.package.sets('set4')._path.exists())
        self.package.delete_set('set4')

    def test_func_delete(self):
        """
        This function tests the delete function of the package class
        """
        self.assertRaises(NoSetFound, self.package.sets, 'set4')

    def test_check_valid(self):
        """
        It checks if the package is valid.
        """
        self.assertTrue(self.package.check_package())

    def test_make_commit(self):
        """
        It tests if the function make_commit() works.
        """
        new_comm = self.package.make_commit('2')
        self.assertIsInstance(new_comm, Package)
        self.assertEqual(new_comm._commit, '2')
        shutil.rmtree(new_comm.commit_path)

    def test_make_commit_and_delete(self):
        """
        It tests if deleting commit works.
        """
        new_comm = self.package.make_commit('3')
        self.assertIsInstance(new_comm, Package)
        self.assertEqual(new_comm._commit, '3')
        path_to_comm = new_comm.commit_path
        new_comm.delete()
        self.assertFalse(path_to_comm.exists())

    def test_prepare_build(self):
        """
        It tests if the function prepare_build() works.
        """
        self.package.prepare_build('test')
        self.assertTrue((self.package.commit_path / '.build' / 'test').is_dir())
        self.package.delete_build('test')
        self.assertFalse((self.package.commit_path / '.build' / 'test').is_dir())

    def test_prepare_build_multiple(self):
        """
        It tests if the function prepare_build() works.
        """
        self.package.prepare_build('test')
        self.package.prepare_build('test2')
        self.assertTrue((self.package.commit_path / '.build' / 'test').is_dir())
        self.package.delete_build()
        self.assertFalse((self.package.commit_path / '.build').is_dir())

    def test_doc_extension_check(self):
        """
        It tests if the doc_extension function works.
        """
        self.assertEqual(self.package.doc_extension('pdf'), 'md')
        self.assertEqual(self.package.doc_extension(), 'md')
        doc_file = self.package.doc_path('md')
        self.assertTrue(doc_file.is_file())

    def test_doc_path(self):
        """
        It tests if the doc_path function works.
        """
        with self.assertRaises(InvalidFileExtension):
            self.package.doc_path('x')
        with self.assertRaises(FileNotFoundError):
            self.package.doc_path('pdf')

    def test_docs_to_display_all(self):
        """
        It tests if the get_docs_to_display function works.
        """
        docs = self.package.get_docs_to_display(best_and_pdf=False)
        self.assertEqual(len(docs), 2)
        for doc in docs:
            self.assertTrue(doc.is_file())
            self.assertTrue(doc.suffix in ('.md', '.txt'))

    def test_docs_to_display_best_and_pdf(self):
        """
        It tests if the get_docs_to_display function works.
        """
        docs = self.package.get_docs_to_display(best_and_pdf=True)
        self.assertEqual(len(docs), 1)
        self.assertTrue(docs[0].is_file())
        self.assertTrue(docs[0].suffix == '.md')


class PackageZip(TestCase):
    def setUp(self):
        self.test_base = Path(__file__).parent / 'test_packages'
        self.zip_path = self.test_base / 'zip_src'
        self.pkg_path = self.test_base / 'zip_pkg'

    def tearDown(self):
        if self.pkg_path.exists():
            shutil.rmtree(self.pkg_path)

    def test_01_create_simple_package(self):
        """
        Tests creating simple package from zip file, checks primary values
        """
        pkg = Package.create_from_zip(self.pkg_path, '1', self.zip_path / 'test_pkg.zip')
        self.assertEqual(pkg['title'], 'zip test pkg')
        self.assertEqual(pkg['points'], 123)

    def test_02_create_simple_package_depth_test(self):
        """
        Tests creating simple package from zip file, checks in-depth values
        """
        pkg = Package.create_from_zip(self.pkg_path, '1', self.zip_path / 'test_pkg.zip')
        self.assertEqual(pkg['title'], 'zip test pkg')

        set1 = pkg.sets('set1')
        self.assertEqual(set1['name'], 'set1')
        self.assertEqual(set1['points'], 10)
        test1 = set1.tests('1')
        with open(test1['input'], 'r') as f:
            self.assertEqual(f.read(), 'set1 in')
        with open(test1['output'], 'r') as f:
            self.assertEqual(f.read(), 'set1 out')



from baca2PackageManager.consts import MEM_SIZES


def bytes_from_str(val: str):
    """
     function is converting memory from others units to bytes

     :return: Memory converted to bytes. (In INT type)
    """
    mem_val = int(val[:-1])
    mem_letter = val[-1]
    mem_val *= MEM_SIZES.get(mem_letter, 1)
    return mem_val


def bytes_to_str(val: int) -> str:
    """
    function is converting memory from bytes to others units

    :param val: Memory in bytes
    :type val: int

    :return: Memory converted to other units. (In STR type)
    :rtype: str
    """

    res_id = 0
    mem_steps = list(MEM_SIZES.items())
    while val // mem_steps[res_id][1] > 0:
        res_id += 1
    if res_id == 1:
        return f"{val}{mem_steps[res_id - 1][0]}"
    elif res_id == 0:
        return f"{val}{mem_steps[res_id][0]}"
    return f"{val / mem_steps[res_id - 1][1]:.2f}{mem_steps[res_id - 1][0]}"



import setuptools
import urllib.request
import json

from pkg_resources import parse_version

MANUAL_VERSION = None


def list_versions(package_name):
    url = f"https://pypi.org/pypi/{package_name}/json"

    with urllib.request.urlopen(url) as res:
        data = json.loads(res.read().decode('utf-8'))

    versions = data['releases']

    return sorted(versions, key=parse_version, reverse=True)


def version(new_version: bool = False) -> str:
    """
    Returns version of the new package, by elevating the version of the old package. Or returns the
    manual version.
    """
    versions = list_versions('baca2-package-manager')
    v = versions[0]
    if len(versions) == 0:
        v = '0.0.1'

    if not new_version:
        return v

    if MANUAL_VERSION is not None:
        mv = MANUAL_VERSION
        if mv not in versions:
            v = mv

    new_v = v.split('.')
    new_v[-1] = str(int(new_v[-1]) + 1)
    new_v = '.'.join(new_v)
    return new_v


setuptools.setup(
    name='baca2-package-manager',
    version=version(new_version=True),
    author='Baca2 Team',
    author_email='bartosz.deptula@student.uj.edu.pl',
    description='A package manager for Baca2 project',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/BaCa2-project/BaCa2-package-manager',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
    ],
    python_requires='>=3.11',
    install_requires=[
        'pyyaml',
        'pydantic'
    ]
)



from hashlib import sha256
import re

import pydantic


class BacaToBroker(pydantic.BaseModel):
    pass_hash: str
    submit_id: str
    package_path: str
    commit_id: str
    submit_path: str


class TestResult(pydantic.BaseModel):
    name: str
    status: str
    time_real: float = 0.0
    time_cpu: float = 0.0
    runtime_memory: int = 0
    answer: str = ''

    logs: dict[str, str] = {}


class SetResult(pydantic.BaseModel):
    name: str
    tests: dict[str, TestResult]


class BrokerToBaca(pydantic.BaseModel):
    pass_hash: str
    submit_id: str
    results: dict[str, SetResult]


class BrokerToBacaError(pydantic.BaseModel):
    pass_hash: str
    submit_id: str
    error_id: str = '0'
    error_data: dict[str, str]


def create_broker_submit_id(course_name: str, submit_id: int) -> str:
    return f'{course_name}___{submit_id}'


def split_broker_submit_id(broker_submit_id: str) -> tuple[str, int]:
    r = re.compile(r'([-A-Za-z0-9.,_]+)___([0-9]+)')
    m = re.fullmatch(r, broker_submit_id)
    course_name = m.group(1)
    submit_id = int(m.group(2))
    return course_name, submit_id


def make_hash(password: str, broker_submit_id: str) -> str:
    hash_obj = sha256()
    hash_obj.update((password + '___').encode('utf-8'))
    hash_obj.update(broker_submit_id.encode('utf-8'))
    return hash_obj.hexdigest()
